"""Compares robustness values

Contains (1) a function for showing how a different set of scenarios
affects the robustness values and robustness rankings; and (2) a
function for showing how different robustness metrics affects the
robustness values and robustness rankings.

Also contains a helper function for creating basic visualisations of
the effects of scenarios and robustness metrics.
"""

import copy
import numpy as np
from scipy import stats
from matplotlib import pyplot as plt
import matplotlib as mpl
from mpl_toolkits.axes_grid1 import make_axes_locatable

def scenarios_similarity(R):
    """Determines similarity in robustness from multiple scenario sets

    Robustness is a function of scenarios, decision alternatives, and
    a performance metric. 2 sets of scenarios can lead to a different
    calculation of robustness. This function measures the difference
    in 2 different ways:
    - Relative difference, delta (%); and
    - Kendall's Tau-b correlation, tau (unitless, [-1, 1]).

    Parameters
    ----------
    R : numpy.ndarray, shape=(m, n)
        Robustness values, R, for m decision alternatives
        and n scenario sets.

    Returns
    -------
    delta : numpy.ndarray, shape=(n, n)
        Average relative difference (%) in robustness for each pair of
        scenario sets. i.e. idx [0, 3] would be the relative difference
        between scenario sets 0 and 3 (and would be equal to [3, 0])
    tau : numpy.ndarray, shape=(n, n)
        Kendall's Tau-b correlation for each pair of scenario sets.
        i.e. idx [0, 3] would be the correlation between scenario
        sets 0 and 3 (and would be equal to [3, 0])
    """
    # Get the number of sets of scenarios
    n = R.shape[1]
    deltas = np.zeros((n, n))
    taus = np.zeros((n, n))

    for idx_1 in range(n):
        for idx_2 in range(idx_1, n):
            delta = np.divide(
                np.abs(R[:, idx_1] - R[:, idx_2]),
                (np.abs(R[:, idx_1] + R[:, idx_2])) / 2.)
            delta = np.average(delta) * 100.0
            deltas[idx_1, idx_2] = delta
            deltas[idx_2, idx_1] = delta
            tau, _ = stats.kendalltau(R[:, idx_1], R[:, idx_2], nan_policy='omit')
            taus[idx_1, idx_2] = tau
            taus[idx_2, idx_1] = tau

    return deltas, taus


def R_metric_similarity(R):
    """Determines similarity in robustness from multiple robustness metrics

    2 different robustness metrics can lead to a different
    calculation of robustness. This function measures the difference
    by using Kendall's Tau-b correlation, tau (unitless, [-1, 1]).

    Parameters
    ----------
    R : numpy.ndarray, shape=(m, n)
        Robustness values, R, for m decision alternatives
        and n robustness metrics.

    Returns
    -------
    tau : numpy.ndarray, shape=(n, n)
        Kendall's Tau-b correlation for each pair of robustness metrics.
        i.e. idx [0, 3] would be the correlation between R metrics
        0 and 3 (and would be equal to [3, 0])
    """
    # Get the number of sets of scenarios
    n = R.shape[1]
    taus = np.zeros((n, n))

    for idx_1 in range(n):
        for idx_2 in range(idx_1, n):
            tau, _ = stats.kendalltau(R[:, idx_1], R[:, idx_2], nan_policy='omit')
            taus[idx_1, idx_2] = tau
            taus[idx_2, idx_1] = tau

    return taus


def delta_plot(delta):
    """A helper fn for plotting the deltas

    Plots the deltas as a 2D heatmap grid.

    Parameters
    ----------
    delta : numpy.ndarray, shape=(n, n)
        Average relative difference (%) in robustness for each pair of
        scenario sets. i.e. idx [0, 3] would be the relative difference
        between scenario sets 0 and 3 (and would be equal to [3, 0])
    """
    ax = plt.subplot(111)
    cmap = 'rainbow'
    cmap = copy.copy(mpl.cm.get_cmap(cmap))
    cmap.set_bad(cmap(0.0))
    norm = mpl.colors.LogNorm(vmin=1.0, vmax=100.0)
    im = ax.imshow(delta, cmap=cmap, norm=norm)
    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="5%", pad=0.2)
    plt.colorbar(im, cax=cax)
    plt.show()


def tau_plot(tau):
    """A helper fn for plotting the Kendall's Tau-b values

    Plots the tau values as a 2D heatmap grid.

    Parameters
    ----------
    tau : numpy.ndarray, shape=(n, n)
        Kendall's Tau-b correlation for each pair of scenario sets.
        i.e. idx [0, 3] would be the correlation between scenario
        sets 0 and 3 (and would be equal to [3, 0])
    """
    ax = plt.subplot(111)
    im = ax.imshow(tau, cmap='RdBu', vmin=-1.0, vmax=1.0)
    divider = make_axes_locatable(ax)
    cax = divider.append_axes("right", size="5%", pad=0.2)
    plt.colorbar(im, cax=cax)
    plt.show()



"""Contains common robustness metrics.

All metrics are explained in detail in:

McPhail, C., Maier, H. R., Kwakkel, J. H., Giuliani, M.,
Castelletti, A., & Westra, S. (2018).
Robustness metrics: How are they calculated, when should they be used
and why do they give different results?. Earth's Future, 6(2), 169-191.
https://doi.org/10.1002/2017EF000649
"""

import numpy as np

from .transforms import t1, t2, t3


def maximin(f, maximise=True):
    """Maximin metric (worst-case scenario)

    The maximin (minimax) metric was first used by Wald (1950).
    It is a very risk averse metric that assumes that the scenario
    that will occur is the scenario under which the performance
    is lowest.

    If minimising, values are made negative so that the aim
    becomes to maximise performance.

    Parameters
    ----------
    f : numpy.ndarray, shape=(m, n)
        Performance values, f, for m decision alternatives
        and n scenarios.
    maximise : bool, optional
        Is the performance metric to be maximised or minimised.
        (The default is True, which implies high values of f are better
        than low values of f).

    Returns
    -------
    numpy.ndarray, shape=(m, )
        The robustness value for each of the m decision alternatives
    """
    _f = t1.identity(f, maximise=maximise)
    _f = t2.worst_case(_f)
    R = t3.f_sum(_f)
    return R


def maximax(f, maximise=True):
    """Maximax metric (best-case scenario)

    Maximax is the opposite of the maximin metric (Wald, 1950). It
    is a metric with a low level of risk aversion that looks for the
    best possible performance that is possible in an decision
    alternative.

    If minimising, values are made negative so that the aim
    becomes to maximise performance.

    Parameters
    ----------
    f : numpy.ndarray, shape=(m, n)
        Performance values, f, for m decision alternatives
        and n scenarios.
    maximise : bool, optional
        Is the performance metric to be maximised or minimised.
        (The default is True, which implies high values of f are better
        than low values of f).

    Returns
    -------
    numpy.ndarray, shape=(m, )
        The robustness value for each of the m decision alternatives
    """
    _f = t1.identity(f, maximise=maximise)
    _f = t2.best_case(_f)
    R = t3.f_sum(_f)
    return R


def hurwicz(f, maximise=True, alpha=0.5):
    """Hurwicz's Optimism-Pessimism Rule

    Hurwicz’s optimism-pessimism rule (Hurwicz, 1953) uses a weighted
    sum of the maximin and maximax metrics that have previously been
    discussed. Like the previous metrics, the Hurwicz’s
    optimism-pessimism rule uses the distribution of performances for
    an individual decision alternative (i.e. it does not compare the
    performances of multiple decision alternatives). It has a
    parameter alpha that determines the relative degree of intrinsic
    risk aversion of the metric where 0<alpha<1 is the weighting of
    the maximin (high level of risk aversion) metric. In other words,
    alpha may be described as the proportion of high to low risk
    aversion for the decision-maker. Being composed of both the
    maximin and maximax metrics brings many of the characteristics of
    these metrics.

    If minimising, values are made negative so that the aim
    becomes to maximise performance.

    Parameters
    ----------
    f : numpy.ndarray, shape=(m, n)
        Performance values, f, for m decision alternatives
        and n scenarios.
    maximise : bool, optional
        Is the performance metric to be maximised or minimised.
        (The default is True, which implies high values of f are better
        than low values of f).
    alpha : float, optional
        The weighting to place on the worst-case scenario.
        (The default is 0.5, which implies an equal weighting of the
        best- and worst-case scenarios).

    Returns
    -------
    numpy.ndarray, shape=(m, )
        The robustness value for each of the m decision alternatives
    """
    # Define the weights for the worst- and best-cases.
    weights = np.asarray([alpha, 1. - alpha])
    _f = t1.identity(f, maximise=maximise)
    _f = t2.worst_and_best_cases(_f)
    R = t3.f_w_sum(_f, weights=weights)
    return R


def laplace(f, maximise=True):
    """Laplace's Principle of Insufficient Reason

    Laplace’s principle of insufficient reason
    (Laplace and Simon, 1951) states that in the absence of information
    on the relative probabilities of the scenarios, each scenario
    should be treated as equally likely. This is equivalent to assuming
    the mean performance across the distributions represents the
    expected value of robustness. Unlike the previously discussed
    metrics, Laplace’s principle of insufficient reason uses the
    performance values from every scenario rather than just using one
    or two performance values.

    If minimising, values are made negative so that the aim
    becomes to maximise performance.

    Parameters
    ----------
    f : numpy.ndarray, shape=(m, n)
        Performance values, f, for m decision alternatives
        and n scenarios.
    maximise : bool, optional
        Is the performance metric to be maximised or minimised.
        (The default is True, which implies high values of f are better
        than low values of f).

    Returns
    -------
    numpy.ndarray, shape=(m, )
        The robustness value for each of the m decision alternatives
    """
    _f = t1.identity(f, maximise=maximise)
    _f = t2.all_scenarios(_f)
    R = t3.f_mean(_f)
    return R


def minimax_regret(f, maximise=True):
    """Minimax Regret metric

    Rather than looking at individual decision alternatives, regret
    metrics including minimax regret (Savage, 1951) look for the
    regret of choosing a particular option. Specifically, minimax
    regret calculates the maximum regret that can be expected in
    any scenario. The regret for a decision alternative l_i in
    scenario s_j is calculated by comparing the performance
    f(x_i, s_j) to the best possible performance of any decision
    alternative in scenario s_j. For decision alternative x_i, the
    robustness value is the regret from the scenario with the
    greatest level of regret. In this case, the objective is to
    minimize the regret. Unlike other metrics which consider an
    individual decision alternative, the minimax regret metric is
    sensitive to the distributions of performance of two or more
    decision alternatives. However, it is only sensitive to the
    largest difference between the distributions.

    If minimising, values are made negative so that the aim
    becomes to maximise performance.

    Parameters
    ----------
    f : numpy.ndarray, shape=(m, n)
        Performance values, f, for m decision alternatives
        and n scenarios.
    maximise : bool, optional
        Is the performance metric to be maximised or minimised.
        (The default is True, which implies high values of f are better
        than low values of f).

    Returns
    -------
    numpy.ndarray, shape=(m, )
        The robustness value for each of the m decision alternatives
    """
    _f = t1.regret_from_best_da(f, maximise=maximise)
    _f = t2.worst_case(_f)
    R = t3.f_sum(_f)
    return R


def percentile_regret(f, maximise=True, percentile=0.1):
    """percentile regret metric

    This is derived from the 90th percentile minimax regret metric
    (Herman et al., 2015) which itself is a variant of the minimax
    metric (Savage, 1951) that was discussed previously. Regret is
    calculated using the same transformation as the minimax regret
    metric, and thus this metric also is used to compare two or more
    decision alternatives rather than only looking at an individual
    decision alternative. The expected amount of regret for decision
    alternative x_i is calculated using the kth percentile of regret
    rather than the maximum possible regret.

    This metric is thus more sensitive to the overall distribution of
    the performance when compared to the minimax regret metric.
    However, it is still most sensitive to only a small number of
    scenarios when compared to a metric such as Laplace’s principle of
    insufficient reason which uses the average of every scenario.


    If minimising, values are made negative so that the aim
    becomes to maximise performance.

    Parameters
    ----------
    f : numpy.ndarray, shape=(m, n)
        Performance values, f, for m decision alternatives
        and n scenarios.
    maximise : bool, optional
        Is the performance metric to be maximised or minimised.
        (The default is True, which implies high values of f are better
        than low values of f).
    percentile : float, optional
        Which percentile of regret values to use.
        (The default is 0.1, which implies the use of the 10th
        percentile. That is the f value at which only 10% of f values
        (for a decision alternative) are worse).

    Returns
    -------
    numpy.ndarray, shape=(m, )
        The robustness value for each of the m decision alternatives
    """
    _f = t1.regret_from_best_da(f, maximise=maximise)
    _f = t2.select_percentiles(_f, np.asarray([percentile]))
    R = t3.f_sum(_f)
    return R


def mean_variance(f, maximise=True):
    """Mean-variance metric

    The mean-variance metric (Kwakkel et al., 2016b) is similar to
    Laplace’s principle of insufficient reason in that it uses the
    mean to determine the expected value of the distribution of
    performances for an individual decision alternative. Unlike
    Laplace’s principle of insufficient reason, the mean-variance
    metric also considers the variability in the distribution of
    performances by using the standard deviation of performance
    values. This metric does face several challenges including
    that the influence of the mean and standard deviation will
    depend on their relative magnitude and thus the trade-off
    between mean and standard deviation is unknown
    (Kwakkel et al., 2016b).

    If minimising, values are made negative so that the aim
    becomes to maximise performance.

    Parameters
    ----------
    f : numpy.ndarray, shape=(m, n)
        Performance values, f, for m decision alternatives
        and n scenarios.
    maximise : bool, optional
        Is the performance metric to be maximised or minimised.
        (The default is True, which implies high values of f are better
        than low values of f).

    Returns
    -------
    numpy.ndarray, shape=(m, )
        The robustness value for each of the m decision alternatives
    """
    _f = t1.identity(f, maximise=maximise)
    _f = t2.all_scenarios(_f)
    R = t3.f_mean_variance(_f)
    return R


def undesirable_deviations(f, maximise=True):
    """Undesirable deviations metric

    The undesirable deviations metric (Kwakkel et al., 2016b) is a
    variation on the approach used by Takriti & Ahmed (2004]). This
    metric only considers undesirable deviations (regret) away from
    the median performance value (which is considered the expected
    value).

    If minimising, values are made negative so that the aim
    becomes to maximise performance.

    Parameters
    ----------
    f : numpy.ndarray, shape=(m, n)
        Performance values, f, for m decision alternatives
        and n scenarios.
    maximise : bool, optional
        Is the performance metric to be maximised or minimised.
        (The default is True, which implies high values of f are better
        than low values of f).

    Returns
    -------
    numpy.ndarray, shape=(m, )
        The robustness value for each of the m decision alternatives
    """
    # Do identity first, before regret, so that correct percentiles
    # can be determined.
    _f = t1.regret_from_median(f, maximise=maximise)
    _f = t2.worst_half(_f)
    R = t3.f_sum(_f)
    return R


def percentile_skew(f, maximise=True):
    """A calculation of skew based on percentiles

    The percentile-based skewness metric (Voudouris et al., 2014)
    considers the skewness of the distribution of performance values.
    This metric gives preference to decision alternatives where the
    performance values are skewed towards better performance values. It
    uses the 10th, 50th and 90th percentile values.

    If minimising, values are made negative so that the aim
    becomes to maximise performance.

    Parameters
    ----------
    f : numpy.ndarray, shape=(m, n)
        Performance values, f, for m decision alternatives
        and n scenarios.
    maximise : bool, optional
        Is the performance metric to be maximised or minimised.
        (The default is True, which implies high values of f are better
        than low values of f).

    Returns
    -------
    numpy.ndarray, shape=(m, )
        The robustness value for each of the m decision alternatives
    """
    _f = t1.identity(f, maximise=maximise)
    # This calculation of skew relies on the 10th, 50th and 90th
    # percentiles.
    percentiles = np.asarray([0.1, 0.5, 0.9])
    _f = t2.select_percentiles(_f, percentiles)
    R = t3.f_skew(_f)
    return R


def percentile_kurtosis(f, maximise=True):
    """A calculation of kurtosis based on percentiles

    A variation of Kurtosis was applied by Voudouris et al. (2014) to
    determine robustness. This metric indicates the “peakedness” of
    the distribution (Kwakkel et al., 2016). It uses the 10th, 25th,
    75th and 90th percentile performance values respectively for each
    decision alternative. Unlike the percentile-based skewness metric,
    this metric does not consider whether the distribution is skewed
    towards higher or lower performance values. A higher value implies
    that the performance values are more peaked around the median
    value.

    Parameters
    ----------
    f : numpy.ndarray, shape=(m, n)
        Performance values, f, for m decision alternatives
        and n scenarios.
    maximise : bool, optional
        Is the performance metric to be maximised or minimised.
        (The default is True, which implies high values of f are better
        than low values of f).

    Returns
    -------
    numpy.ndarray, shape=(m, )
        The robustness value for each of the m decision alternatives
    """
    _f = t1.identity(f, maximise=maximise)
    # This calculation of skew relies on the 10th, 50th and 90th
    # percentiles.
    percentiles = np.asarray([0.1, 0.25, 0.75, 0.9])
    _f = t2.select_percentiles(_f, percentiles)
    R = t3.f_kurtosis(_f)
    return R


def starrs_domain(f, maximise=True, threshold=0.0, accept_equal=True):
    """Robustness based on proportion of scenarios meeting a threshold

    Unlike previous metrics, Starr’s domain criterion (Starr, 1963;
    Schneller and Sphicas, 1983) compares the distribution of
    performance values to a threshold value. This metric is most useful
    when the threshold is selected such that the level of performance
    above or below the threshold does not matter, but preferably the
    decision alternative will meet this threshold. For example, a
    system may have a threshold such that any decision alternative
    with a performance below the threshold is a fail and any
    performance above the threshold is a pass.

    Parameters
    ----------
    f : numpy.ndarray, shape=(m, n)
        Performance values, f, for m decision alternatives
        and n scenarios.
    maximise : bool, optional
        Is the performance metric to be maximised or minimised.
        (The default is True, which implies high values of f are better
        than low values of f).
    threshold : float, optional
        A minimum value where f >= threshold to be satisficed
        (The default is 0.0, which implies that any f value above 0 is
        of satisfactory performance).
    accept_equal : bool, optional
        Whether or not an f value equal to the threshold is acceptable.
        (The default is True, which implies a >= comparison, whereas
        False would imply a > comparison).

    Returns
    -------
    numpy.ndarray, shape=(m, )
        The robustness value for each of the m decision alternatives
    """
    _f = t1.satisfice(
        f,
        maximise=maximise,
        threshold=threshold,
        accept_equal=accept_equal)
    _f = t2.all_scenarios(_f)
    R = t3.f_mean(_f)
    return R



"""Contains T3 performance transformations (aggregation to robustness)

Generally the selected and transformed performance values, f, are
transformed to an expected value of performance.

However, supplementary metrics may consider the variance in f,
or higher-order moments of f.
"""

import numpy as np


def f_identity(f):
    """Identity transform included for completeness.

    Parameters
    ----------
    f : np.ndarray, shape=(m, 1)
        Transformed performance values to be maximised.
        m decision alternatives and n scenarios

    Returns
    -------
    np.ndarray, shape=(m, )
        The robustness value for each of the m decision alternatives
    """
    R = np.reshape(f, newshape=(-1, ))
    return R


def f_mean(f):
    """Calculate robustness as mean of f

    Parameters
    ----------
    f : np.ndarray, shape=(m, n)
        Transformed performance values to be maximised.
        m decision alternatives and n scenarios

    Returns
    -------
    np.ndarray, shape=(m, )
        The robustness value for each of the m decision alternatives
    """
    R = np.mean(f, axis=1)
    return R


def f_range(f):
    """Calculate robustness as range of f

    Parameters
    ----------
    f : np.ndarray, shape=(m, n)
        Transformed performance values to be maximised.
        m decision alternatives and n scenarios

    Returns
    -------
    np.ndarray, shape=(m, )
        The robustness value for each of the m decision alternatives
    """
    max_f = np.max(f, axis=1)
    min_f = np.min(f, axis=1)
    R = max_f - min_f
    return R


def f_sum(f):
    """Calculate robustness as sum of f (/ n_scenarios)

    Parameters
    ----------
    f : np.ndarray, shape=(m, n)
        Transformed performance values to be maximised.
        m decision alternatives and n scenarios

    Returns
    -------
    np.ndarray, shape=(m, )
        The robustness value for each of the m decision alternatives
    """
    R = np.sum(f, axis=1) / f.shape[1]
    return R


def f_w_sum(f, weights):
    """Calculate robustness as weighted sum of f

    Parameters
    ----------
    f : np.ndarray, shape=(m, n)
        Transformed performance values to be maximised.
        m decision alternatives and n scenarios
    weights : np.ndarray, shape=(n, )
        Weights to apply to each scenario
        E.g. for n=3, you could use
        weights = [0.5, 0.25, 0.25]

    Returns
    -------
    np.ndarray, shape=(m, )
        The robustness value for each of the m decision alternatives
    """
    R = np.matmul(f, weights)
    return R


def f_variance(f):
    """Calculate robustness as variance of f

    Parameters
    ----------
    f : np.ndarray, shape=(m, n)
        Transformed performance values to be maximised.
        m decision alternatives and n scenarios

    Returns
    -------
    np.ndarray, shape=(m, )
        The robustness value for each of the m decision alternatives
    """
    # Calculate variance with ddof=1
    # (variance of sample, not population)
    R = np.var(f, axis=1, ddof=1)
    return R


def f_mean_variance(f):
    """Calculate robustness as a combination of mean and variance of f

    Parameters
    ----------
    f : np.ndarray, shape=(m, n)
        Transformed performance values to be maximised.
        m decision alternatives and n scenarios

    Returns
    -------
    np.ndarray, shape=(m, )
        The robustness value for each of the m decision alternatives
    """
    mean_f = f_mean(f)
    # Calculate variance with ddof=1
    # (std deviation of sample, not population)
    std_dev_f = np.std(f, axis=1, ddof=1)
    # +1 is to ensure no divide by 0
    R = np.divide((mean_f + 1), (std_dev_f + 1))
    return R


def f_skew(f, reverse=False):
    """Calculate robustness as the skew of f

    It assumes that it is best to have most values skewed towards
    high performance, with a larger tail for low performance.

    Parameters
    ----------
    f : np.ndarray, shape=(m, 3)
        Transformed performance values to be maximised.
        m decision alternatives and 3 scenarios.
        Those 3 scenarios must be (in order) the 10th, 50th and 90th
        percentiles, where the 10th percentile, p10, is f where only
        10% of f is worse than p10. If wanting the reverse of this,
        see arg 'reverse' below.
    reverse : bool, optional
        Reverses the skew calculation to have a preference for a
        larger tail of higher-performance values.
        (The default is False, which implies it is best to have a skew
        towards high f with a larger tail towards low f).

    Returns
    -------
    np.ndarray, shape=(m, )
        The robustness value for each of the m decision alternatives
    """
    # p10 < p50 < p90
    p10 = f[:, 0]
    p50 = f[:, 1]
    p90 = f[:, 2]
    # If p50 is closer to p90 than p10, most of the values are
    # skewed towards the higher-performance end of f
    # i.e. If p50 > mu50, most values of f are skewed higher
    if not reverse:
        R = p50 - ((p90 + p10) / 2.)
    else:
        R = ((p90 + p10) / 2.) - p50
    normaliser = (p90 - p10) / 2.
    R = np.divide(R, normaliser)
    return R


def f_kurtosis(f):
    """Calculate robustness as the kurtosis of f

    Parameters
    ----------
    f : np.ndarray, shape=(m, 4)
        Transformed performance values.
        m decision alternatives and 4 scenarios.
        Those 4 scenarios must be (in order) the 10th, 25th, 75th and
        90th percentiles, where the 10th percentile, q10, is f where
        only 10% of f is worse than q10.

    Returns
    -------
    np.ndarray, shape=(m, )
        The robustness value for each of the m decision alternatives
    """
    # p10 < p25 < p75 < p90
    p10 = f[:, 0]
    p25 = f[:, 1]
    p75 = f[:, 2]
    p90 = f[:, 3]
    R = np.divide((p90 - p10), (p75 - p25))
    return R



"""Helper functions for calculating robustness from performance values"""

import numpy as np


def f_to_R(f_df, R_dict):
    """Calculates robustness from performance values.

    Uses a set of performance values, `f`, determined from simulations
    across multiple decision alternatives, `l`, different scenarios,
    `s`, and calculates robustness, `R`, using a variety of given
    robustness metrics.

    Parameters
    ----------
    f_df : pandas.DataFrame
        A dataframe of performance values, `f`, with indexes for the
        scenario, `s`, and decision alternative, `l`.
        Must include any other variables to be used during calculation
        of robustness (referred to as 'varX_name' below).
        Columns: `['s_idx', 'l_idx', '<f1_name>', '<f2_name>', ...,
                   '<var1_name>', '<var2_name>', ...]`
    R_dict : dict of dict
        A mapping of robustness metric (`R`) names to information
        about those robustness metrics including
        'f': string
            the corresponding performance metric to use
        'maximise': bool
            whether the aim of that performance metric is to be maximised
        'threshold': None or string
            the name of the column in `f_df` containing thresholds
            OR None if not using a threshold or threshold is given in kwargs
        'func': func
            the robustness metric function
        'kwargs': dict
            keyword arguments required for calculating R
            e.g. {'t1_kwargs': {'threshold': 5.2}}
                 would pass a threshold kwarg to the t1 transformation
                 if using the custom_R_metric
        Note that all performance metric names must be listed here.
        E.g. `{'<R1_name>': {'f': <f1_name>, 'maximise': <bool>, 'threshold': None, 'func': <func>, 'kwargs': {'kwarg1': <arg>}},
               '<R2_name>': {'f': <f1_name>, 'maximise': <bool>, 'threshold': 'critical', 'func': <func>, 'kwargs': {}}, ...}`

    Returns
    -------
    pandas.DataFrame
        A dataframe of robustness values, `R`, with indexes for the
        decision alternative, `l`, and a column for the performance
        metric name, ``f_name``.
        Columns: `['l_idx', 'f_name', '<R1_name>', '<R2_name>', ...]`
    """
    # Extract the names of the f metrics
    f_metrics = [R_dict[R_name]['f'] for R_name in R_dict]
    df_cols = [col for col in f_df.columns]
    # Check that the same metrics are in f_maximise
    for f_metric in f_metrics:
        assert f_metric in df_cols

    # Sort dataframe to ensure consistency
    f_df = sort_f_df(f_df)

    # Get the scenario and decision alternative idxs, and
    # check that the s_idx and l_idx indexes are valid.
    s_idxs, l_idxs = get_f_df_details(f_df)

    # Loop through performance metrics
    R = {}
    for R_metric in R_dict:
        f_metric = R_dict[R_metric]['f']
        # Check that required data exists
        kwargs = R_dict[R_metric]['kwargs']
        if R_dict[R_metric]['threshold'] is not None:
            assert R_dict[R_metric]['threshold'] in df_cols
            if 't1_kwargs' not in kwargs:
                kwargs['t1_kwargs'] = {}
            kwargs['t1_kwargs']['threshold'] = np.reshape(
                f_df.iloc[:, f_df.columns.get_loc(R_dict[R_metric]['threshold'])].values,
                newshape=(l_idxs.size, s_idxs.size))
        f = np.reshape(
            f_df.iloc[:, f_df.columns.get_loc(f_metric)].values,
            newshape=(l_idxs.size, s_idxs.size))
        kwargs['maximise'] = R_dict[R_metric]['maximise']
        R[R_metric] = R_dict[R_metric]['func'](f, **kwargs)
    return R


def sort_f_df(f_df):
    """Sorts f_df by s_idx first then by l_idx.

    E.g. for scenario 0, see all decision alternatives in order,
    then scenario 1, scenario 2, etc.

    Parameters
    ----------
    f_df : pandas.DataFrame
        A dataframe of performance values, `f`, with indexes for the
        scenario, `s`, and decision alternative, `l`.
        Columns: `['s_idx', 'l_idx', '<f1_name>', '<f2_name>', ...]`
    """
    # This will sort first by s_idx then by l_idx, both from 0 to ...
    f_df.sort_values(['l_idx', 's_idx'], ascending=[True, True])
    return f_df


def get_f_df_details(f_df):
    """Gets the unique s_idx and l_idx values in f_df.

    Also checks that for each s_idx, each unique l_idx exists
    (and vice versa).

    Parameters
    ----------
    f_df : pandas.DataFrame
        A dataframe of performance values, `f`, with indexes for the
        scenario, `s`, and decision alternative, `l`.
        Columns: `['s_idx', 'l_idx', '<f1_name>', '<f2_name>', ...]`

    Returns
    -------
    s_idxs : list of int
        Number of scenario (`s`) idxs
    l_idxs : list of int
        List decision alternative (`l`) idxs
    """
    # Check that df is sorted
    f_df = sort_f_df(f_df)

    s_idxs = f_df['s_idx'].unique()
    l_idxs = f_df['l_idx'].unique()

    for s_idx in s_idxs:
        relevant_rows = f_df.loc[f_df['s_idx'] == s_idx]
        relevant_l_idxs = relevant_rows['l_idx'].values
        assert np.allclose(relevant_l_idxs, l_idxs)

    return s_idxs, l_idxs



"""Tests the T2 transformations."""

import numpy as np
from .. import t2


def test_all_scenarios():
    """Tests the all_scenarios fn"""
    f = np.asarray([
        [0.99, 1.0, 0.5],
        [0.69, 0.6, 0.6]])
    expected = np.copy(f)
    _f = t2.all_scenarios(f)
    assert np.allclose(_f, expected)


def test_worst_case():
    """Tests the worst_case fn"""
    f = np.asarray([
        [0.99, 1.0, 0.5],
        [0.69, 0.6, 0.6]])
    _f = t2.worst_case(f)
    expected = np.asarray([
        [0.5],
        [0.6]])
    assert np.allclose(_f, expected)


def test_best_case():
    """Tests the best_case fn"""
    f = np.asarray([
        [0.99, 1.0, 0.5],
        [0.69, 0.6, 0.6]])
    _f = t2.best_case(f)
    expected = np.asarray([
        [1.0],
        [0.69]])
    assert np.allclose(_f, expected)


def test_worst_and_best_cases():
    """Tests the worst_and_best_cases fn"""
    f = np.asarray([
        [0.99, 1.0, 0.5],
        [0.69, 0.6, 0.6]])
    _f = t2.worst_and_best_cases(f)
    expected = np.asarray([
        [0.5, 1.0],
        [0.6, 0.69]])
    assert np.allclose(_f, expected)


def test_worst_half():
    """Tests the worst_half fn"""
    f = np.asarray([
        [0.99, 1.0, 0.5],
        [0.69, 0.6, 0.6]])
    _f = t2.worst_half(f)
    expected = np.asarray([
        [0.5, 0.99],
        [0.6, 0.6]])
    assert np.allclose(_f, expected)


def percentiles():
    """Tests the worst_half fn"""
    f = np.asarray([
        [0.99, 1.0, 0.5, 0.2],
        [0.69, 0.6, 0.6, 0.2]])
    percentiles = np.asarray(
        [0.25, 0.99])
    _f = t2.select_percentiles(f, percentiles)
    expected = np.asarray([
        [0.5, 1.0],
        [0.6, 0.69]])
    assert np.allclose(_f, expected)



"""Contains the T2 transformations (scenario subset selection).

This is related to the level of risk averseness of the decision-maker.
"""

import numpy as np


def all_scenarios(f):
    """Use all scenarios. Provided for completeness.

    Parameters
    ----------
    f : np.ndarray, shape=(m, n)
        Transformed performance values to be maximised.
        m decision alternatives and n scenarios

    Returns
    -------
    np.ndarray, shape=(m, n')
        The selected n' performance values
        In this case n' = n
    """
    return f


def worst_case(f):
    """Assume the worst-case scenario for each decision alternative.

    Parameters
    ----------
    f : np.ndarray, shape=(m, n)
        Transformed performance values to be maximised.
        m decision alternatives and n scenarios

    Returns
    -------
    np.ndarray, shape=(m, n')
        The selected n' performance values
        In this case n' = 1
    """
    worst_f = np.amin(f, 1, keepdims=True)
    return worst_f


def best_case(f):
    """Assume the best-case scenario for each decision alternative.

    Parameters
    ----------
    f : np.ndarray, shape=(m, n)
        Transformed performance values to be maximised.
        m decision alternatives and n scenarios

    Returns
    -------
    np.ndarray, shape=(m, n')
        The selected n' performance values
        In this case n' = 1
    """
    best_f = np.amax(f, 1, keepdims=True)
    return best_f


def worst_and_best_cases(f):
    """Work with the most extreme worst- and best-case scenarios.

    Parameters
    ----------
    f : np.ndarray, shape=(m, n)
        Transformed performance values to be maximised.
        m decision alternatives and n scenarios

    Returns
    -------
    np.ndarray, shape=(m, n')
        The selected n' performance values
        In this case n' = 2
    """
    worst_f = worst_case(f)
    best_f = best_case(f)
    _f = np.concatenate((worst_f, best_f), axis=1)
    return _f


def worst_half(f):
    """Work with the worst half of scenarios

    Parameters
    ----------
    f : np.ndarray, shape=(m, n)
        Transformed performance values to be maximised.
        m decision alternatives and n scenarios

    Returns
    -------
    np.ndarray, shape=(m, n')
        The selected n' performance values
        In this case n' = 0.5*n (round up to nearest whole number)
    """
    sorted_f = np.sort(f)
    n = sorted_f.shape[1]  # Num of scenarios
    _n = int(n / 2. + 0.51)  # Half of the scenarios
    _f = sorted_f[:, :_n]
    return _f


def select_percentiles(f, percentiles):
    """Select particular percentiles of f for each decision alternative.

    Parameters
    ----------
    f : np.ndarray, shape=(m, n)
        Transformed performance values to be maximised.
        m decision alternatives and n scenarios
    percentiles : np.ndarray, shape=(n', ), dtype=float
        Which percentile of to select for each decision alternative.
        E.g. [0.2, 0.75] would get the 20th and 75th percentiles for
        each decision alternative. That is to say, the f values for
        each decision alternative where 20% and 75% of values are
        worse.

    Returns
    -------
    np.ndarray, shape=(m, n')
        The selected n' performance values
        n' is given by the percentiles parameter
    """
    _f = np.transpose(
        np.quantile(f, percentiles, axis=1, interpolation='nearest'))
    return _f



"""Robustness metrics and transforms."""
from .common_metrics import (
    maximin,
    maximax,
    hurwicz,
    laplace,
    minimax_regret,
    percentile_regret,
    mean_variance,
    undesirable_deviations,
    percentile_skew,
    percentile_kurtosis,
    starrs_domain)
from .custom_metrics import custom_R_metric, guidance_to_R
from .transforms import t1, t2, t3



"""Contains the T1 functions (performance value transformations).

In general, T1 will take one of 3 forms:
    1. Identity transform - for understanding actual performance
    2. Regret transform - for understanding cost of making wrong decision
    3. Satisficing transform - for understanding how often constraints
       are satisfied.
It is expected that after T1, the aim is to maximise performance.
i.e. Even for the identity transform, if minimising, then values
will be returned as negative.
"""

import numpy as np


def identity(f, maximise=True):
    """Keeps values the same unless minimising.

    If minimising, values are made negative so that the aim
    becomes to maximise performance.

    Parameters
    ----------
    f : np.ndarray, shape=(m, n)
        Performance values, f, for m decision alternatives
        and n scenarios.
    maximise : bool
        Is the performance metric to be maximised or minimised.
        (The default is True, which implies high values of f are better
        than low values of f).

    Returns
    -------
    np.ndarray, shape=(m, n)
        Transformed performance values, f', for m decision alternatives
        and n scenarios
    """
    _f = _prepare_f(f)
    return _f if maximise else -_f


def regret_from_best_da(f, maximise=True):
    """T1: Regret from best decision alternative

    Returns negative regret, so that from this point on,
    the aim is to maximise the negative regret (towards 0).

    Parameters
    ----------
    f : np.ndarray, shape=(m, n)
        Performance values, f, for m decision alternatives
        and n scenarios.
    maximise : bool
        Is the performance metric to be maximised or minimised.
        (The default is True, which implies high values of f are better
        than low values of f).

    Returns
    -------
    np.ndarray, shape=(m, n)
        Transformed performance values, f', for m decision alternatives
        and n scenarios
    """
    _f = _prepare_f(f)
    _f = identity(_f, maximise=maximise)
    best_decision_alternatives = np.amax(_f, axis=0)
    regret = _f - best_decision_alternatives
    return regret


def satisficing_regret(f, threshold, maximise=True):
    """T1: Satisficing regret

    For a given decision alternative, this function compares its performance in
    each scenario to a threshold (threshold can be different for each decision
    alternative), and calculates the magnitude from the threshold (if it fails)
    or returns 0 if it meets the threshold
    Returns negative regret, so that from this point on,
    the aim is to maximise the negative regret (towards 0).

    Parameters
    ----------
    f : np.ndarray, shape=(m, n)
        Performance values, f, for m decision alternatives
        and n scenarios.
    threshold : np.ndarray, shape=(n, ) or float
        The values to compare the performance values to. i.e. The
        values you would regret not getting, relative to f.
        Can be a different value for each scenario or one value
        across all scenarios.
    maximise : bool
        Is the performance metric to be maximised or minimised.
        (The default is True, which implies high values of f are better
        than low values of f).

    Returns
    -------
    np.ndarray, shape=(m, n)
        Transformed performance values, f', for m decision alternatives
        and n scenarios
    """
    _f = _prepare_f(f)
    # Determine the regret for each solution in each scenario
    # This will be a number with the aim to be maximised.
    # If the solution has better performance than the threshold,
    # then it will be a positive number
    regret = regret_from_values(_f, threshold, maximise=maximise)
    # In satisficing regret, we only care about the magnitude of
    # failure IF there is a failure. So any performances that are
    # not failures are zeroed out.
    regret[regret > 0.] = 0.
    return regret


def regret_from_values(f, values, maximise=True):
    """T1: Regret from given values

    For a given decision alternative, this function compares its performance in
    each scenario to given performance value for that decision alternative
    Returns negative regret, so that from this point on,
    the aim is to maximise the negative regret (towards 0).

    Parameters
    ----------
    f : np.ndarray, shape=(m, n)
        Performance values, f, for m decision alternatives
        and n scenarios.
    values : np.ndarray, shape=(n, ) or float
        The values to compare the performance values to. i.e. The
        values you would regret not getting, relative to f.
        Can be a different value for each scenario or one value
        across all scenarios.
    maximise : bool
        Is the performance metric to be maximised or minimised.
        (The default is True, which implies high values of f are better
        than low values of f).

    Returns
    -------
    np.ndarray, shape=(m, n)
        Transformed performance values, f', for m decision alternatives
        and n scenarios
    """
    _f = _prepare_f(f)
    if isinstance(values, np.ndarray):
        v = values
    else:
        # If values is not different for each scenario, then we must
        # put it in the form of one value, repeated
        v = np.repeat(values, _f.shape[1])
    regret = np.subtract(_f, v)
    # Take into account whether f is to be minimised or maximised.
    regret = identity(regret, maximise=maximise)
    return regret


def regret_from_median(f, maximise=True):
    """T1: Regret from median values

    For a given decision alternative, this function compares its performance in
    each scenario to median performance for that decision alternative

    For a given decision alternative, this function compares its performance in
    each scenario to given performance value for that decision alternative
    Returns negative regret, so that from this point on,
    the aim is to maximise the negative regret (towards 0).

    Parameters
    ----------
    f : np.ndarray, shape=(m, n)
        Performance values, f, for m decision alternatives
        and n scenarios.
    maximise : bool
        Is the performance metric to be maximised or minimised.
        (The default is True, which implies high values of f are better
        than low values of f).

    Returns
    -------
    np.ndarray, shape=(m, n)
        Transformed performance values, f', for m decision alternatives
        and n scenarios
    """
    _f = _prepare_f(f)
    median_f = np.median(_f, axis=1, keepdims=True)
    regret = np.subtract(_f, median_f)
    regret = identity(regret, maximise=maximise)
    return regret


def satisfice(f, maximise=True, threshold=0.0, accept_equal=True):
    """Transform performance how many scenarios are satisficed

    Parameters
    ----------
    f : np.ndarray, shape=(m, n)
        Performance values, f, for m decision alternatives
        and n scenarios.
    maximise : bool
        Is the performance metric to be maximised or minimised.
        (The default is True, which implies high values of f are better
        than low values of f).
    threshold : float, optional
        A minimum value where f >= threshold to be satisficed
        (The default is 0.0, which implies that any f value above 0 is
        of satisfactory performance).
    accept_equal : bool, optional
        Whether or not an f value equal to the threshold is acceptable.
        (The default is True, which implies a >= comparison, whereas
        False would imply a > comparison).

    Returns
    -------
    np.ndarray, shape=(m, n)
        Transformed performance values, f', for m decision alternatives
        and n scenarios
    """
    _f = _prepare_f(f)
    _f = identity(_f, maximise=maximise)
    c = threshold if maximise else -threshold
    _f = _f >= c if accept_equal else _f > c
    _f = np.where(_f, 1., 0.)
    return _f


def _prepare_f(f):
    """Ensures f is in the right form for t1 transformations.

    Converts to an np.ndarray if it isn't already.
    If is of shape (n, ), it converts it to shape (1, n)
    Parameters
    ----------
    f : np.ndarray, shape=(m, n)
        Performance values, f, for m decision alternatives
        and n scenarios.

    Returns
    -------
    np.ndarray, shape=(m, n)
        Performance values, f, for m decision alternatives
        and n scenarios.
    """
    _f = f if isinstance(f, np.ndarray) else np.asarray(f)
    if len(_f.shape) != 2:
        _f = np.reshape(_f, newshape=(1, -1))
    return _f



"""Tests the T1 transformations"""

import numpy as np
from .. import t1


def test_identity():
    """Tests the regret_from_best_da fn"""
    f = np.asarray([
        [0.99, 1.0, 0.5],
        [0.69, 0.6, 0.6]])
    expected = np.copy(f)
    _f = t1.identity(f, maximise=True)
    assert np.allclose(_f, expected)
    _f = t1.identity(f, maximise=False)
    expected = -np.copy(f)
    assert np.allclose(_f, expected)


def test_regret_from_best_da():
    """Tests the regret_from_best_da fn"""
    f = np.asarray([
        [0.99, 1.0, 0.5],
        [0.69, 0.6, 0.6]])
    regret = t1.regret_from_best_da(f, maximise=True)
    expected = np.asarray([
        [0.0, 0.0, -0.1],
        [-0.3, -0.4, 0.0]])
    assert np.allclose(regret, expected)
    regret = t1.regret_from_best_da(f, maximise=False)
    expected = np.asarray([
        [-0.3, -0.4, 0.0],
        [0.0, 0.0, -0.1]])
    assert np.allclose(regret, expected)


def test_regret_from_values():
    """Tests the regret_from_values fn"""
    # Test for array of values
    f = np.asarray([
        [0.99, 1.0, 0.5],
        [0.69, 0.6, 0.6]])
    values = np.asarray(
        [0.79, 1.0, 1.0])
    regret = t1.regret_from_values(f, values, maximise=True)
    expected = np.asarray([
        [0.2, 0.0, -0.5],
        [-0.1, -0.4, -0.4]])
    assert np.allclose(regret, expected)
    regret = t1.regret_from_values(f, values, maximise=False)
    expected = np.asarray([
        [-0.2, 0.0, 0.5],
        [0.1, 0.4, 0.4]])
    assert np.allclose(regret, expected)
    # Test for a single value
    f = np.asarray([
        [0.9, 1.0, 0.5],
        [1.4, 0.6, 0.6]])
    values = 0.9
    regret = t1.regret_from_values(f, values, maximise=True)
    expected = np.asarray([
        [0.0, 0.1, -0.4],
        [0.5, -0.3, -0.3]])
    assert np.allclose(regret, expected)
    regret = t1.regret_from_values(f, values, maximise=False)
    expected = np.asarray([
        [0.0, -0.1, 0.4],
        [-0.5, 0.3, 0.3]])
    assert np.allclose(regret, expected)


def test_satisficing_regret():
    """Tests the satisficing_regret fn"""
    # Test for array of thresholds
    f = np.asarray([
        [0.99, 1.0, 0.5],
        [0.69, 0.6, 0.6]])
    thresholds = np.asarray(
        [0.79, 1.0, 1.0])
    regret = t1.satisficing_regret(f, thresholds, maximise=True)
    expected = np.asarray([
        [0.0, 0.0, -0.5],
        [-0.1, -0.4, -0.4]])
    assert np.allclose(regret, expected)
    regret = t1.satisficing_regret(f, thresholds, maximise=False)
    expected = np.asarray([
        [-0.2, 0.0, 0.0],
        [0.0, 0.0, 0.0]])
    assert np.allclose(regret, expected)
    # Test for a single threshold
    f = np.asarray([
        [0.9, 1.0, 0.5],
        [1.4, 0.6, 0.6]])
    thresholds = 0.9
    regret = t1.satisficing_regret(f, thresholds, maximise=True)
    expected = np.asarray([
        [0.0, 0.0, -0.4],
        [0.0, -0.3, -0.3]])
    assert np.allclose(regret, expected)
    regret = t1.satisficing_regret(f, thresholds, maximise=False)
    expected = np.asarray([
        [0.0, -0.1, 0.0],
        [-0.5, 0.0, 0.0]])
    assert np.allclose(regret, expected)


def test_regret_from_median():
    """Tests the regret_from_median fn"""
    f = np.asarray([
        [0.99, 1.0, 0.5],
        [0.69, 0.6, 0.6]])
    regret = t1.regret_from_median(f, maximise=True)
    expected = np.asarray([
        [0.0, 0.01, -0.49],
        [0.09, 0.0, 0.0]])
    assert np.allclose(regret, expected)
    regret = t1.regret_from_median(f, maximise=False)
    expected = np.asarray([
        [0.0, -0.01, 0.49],
        [-0.09, 0.0, 0.0]])
    assert np.allclose(regret, expected)


def test_satisfice():
    """Tests the f_satisfice fn"""
    f = np.asarray([
        [0.5, 0.99, 1.0],
        [0.6, 0.65, 0.69]])
    threshold = 0.65
    _f = t1.satisfice(
        f,
        maximise=True,
        threshold=threshold,
        accept_equal=True)
    expected = np.asarray([
        [0.0, 1.0, 1.0],
        [0.0, 1.0, 1.0]])
    assert np.allclose(_f, expected)
    _f = t1.satisfice(
        f,
        maximise=True,
        threshold=threshold,
        accept_equal=False)
    expected = np.asarray([
        [0.0, 1.0, 1.0],
        [0.0, 0.0, 1.0]])
    assert np.allclose(_f, expected)
    _f = t1.satisfice(
        f,
        maximise=False,
        threshold=threshold,
        accept_equal=True)
    expected = np.asarray([
        [1.0, 0.0, 0.0],
        [1.0, 1.0, 0.0]])
    assert np.allclose(_f, expected)
    _f = t1.satisfice(
        f,
        maximise=False,
        threshold=threshold,
        accept_equal=False)
    expected = np.asarray([
        [1.0, 0.0, 0.0],
        [1.0, 0.0, 0.0]])
    assert np.allclose(_f, expected)



"""Contains helper code for creating custom robustness metrics.

Guidance for creating custom robustness metrics contained in:
TODO

"""

from .transforms import t1, t2, t3


class custom_R_metric:
    """Create a custom robustness metric

    """
    def __init__(self, t1_func, t2_func, t3_func):
        """Initialize the custom Robustness metric
        """
        self.t1_func = t1_func
        self.t2_func = t2_func
        self.t3_func = t3_func

    def __call__(
            self,
            f,
            maximise=True,
            t1_kwargs=None,
            t2_kwargs=None,
            t3_kwargs=None):
        """Calculate robustness from given values

        Parameters
        ----------
        f : numpy.ndarray, shape=(m, n)
            Performance values, f, for m decision alternatives
            and n scenarios.
        maximise : bool, optional
            Is the performance metric to be maximised or minimised.
            (The default is True, which implies high values of f are better
            than low values of f).
        t1_kwargs, t2_kwargs, t3_kwargs : dict, optional
            The keyword arguments required for these transfromations

        Returns
        -------
        numpy.ndarray, shape=(m, ) OR float if m=1
            The robustness value for each of the m decision alternatives
        """
        if t1_kwargs is None:
            t1_kwargs = {}
        if t2_kwargs is None:
            t2_kwargs = {}
        if t3_kwargs is None:
            t3_kwargs = {}
        transformed_f = self.t1_func(f, maximise=maximise, **t1_kwargs)
        selected_f = self.t2_func(transformed_f, **t2_kwargs)
        R = self.t3_func(selected_f, **t3_kwargs)
        if R.shape[0] == 1:
            R = R[0]
        return R


def callable_transformation(transformation, kwargs):
    """Allows kwargs to be given to transformation before calling it."""
    func = lambda f: transformation(f, **kwargs)
    return func


def guidance_to_R():
    """Guides the user to produce a custom robustness metric.

    Returns
    -------
    class
        A callable class that is the custom robustness metric.
    """
    print('\n\n******')
    print('Create a custom robustness metric')
    print('******\n')
    t1_func = None
    t2_func = None
    t3_func = None

    threshold = ''
    while threshold not in ['y', 'n']:
        print('Does a meaningful threshold for the level of performance exist? (y/n)')
        print('\tE.g. supply must be greater than demand, or')
        print('\t     cost must be kept within a budget?')
        threshold = input()
        if threshold not in ['y', 'n']:
            print('\nError: answer must be "y" or "n".')

    most_important = ''
    if threshold == 'n':
        while most_important not in ['a', 'b']:
            print('\nIs it most important to (a) make the best decision, or (b) avoid making the wrong decision? (a or b)')
            most_important = input()
            if most_important not in ['a', 'b']:
                print('\nError: answer must be "a" or "b".')
        if most_important == 'a':
            t1_func = t1.identity
        else:
            t1_func = t1.regret_from_best_da
    else:
        print('\n** Make sure that the threshold is included in t1_kwargs when calculating robustness! **')
        while most_important not in ['a', 'b']:
            print('\nIs it most important to (a) minimise the magnitude of failure, or (b) maximise number of scenarios with acceptable performance? (a or b)')
            most_important = input()
            if most_important not in ['a', 'b']:
                print('\nError: answer must be "a" or "b".')
        if most_important == 'a':
            t1_func = t1.satisficing_regret
        else:
            t1_func = t1.satisfice
            t3_func = t3.f_mean

    indication_of_f = ''
    if t3_func is None:
        while indication_of_f not in ['a', 'b']:
            print('\nIs it most important to (a) get an indication of the level of performance or (b) the range of performance? (a or b)')
            indication_of_f = input()
            if indication_of_f not in ['a', 'b']:
                print('\nError: answer must be "a" or "b".')
        if indication_of_f == 'a':
            t3_func = t3.f_mean
        else:
            t3_func = t3.f_range

    if indication_of_f in ['', 'b']:
        print('\nSelect an upper and lower percentile to reflect the level of risk aversion/tolerance.')
        print('(i.e. between 0%% and 100%% reflecting maximum risk aversion and maximum risk tolerance, respectively).')
        print('First, enter the upper percentile (e.g. enter "87" for 87%):')
        upper_bnd = float(input()) / 100.
        print('\nNow enter the lower bound:')
        lower_bnd = float(input()) / 100.
        t2_func = callable_transformation(t2.select_percentiles, {'percentiles': [lower_bnd, upper_bnd]})
    else:
        print('\nSelect an percentile to reflect the level of risk aversion/tolerance.')
        print('(i.e. between 0% and 100% reflecting maximum risk aversion and maximum risk tolerance, respectively):')
        percentile = float(input()) / 100.
        t2_func = callable_transformation(t2.select_percentiles, {'percentiles': [percentile]})

    R_metric = custom_R_metric(t1_func, t2_func, t3_func)
    return R_metric



"""Tests the T3 transformations."""

import numpy as np
from .. import t3


def test_f_mean():
    """Tests the f_mean fn"""
    f = np.asarray([
        [0.99, 1.0, 0.5],
        [0.69, 0.6, 0.6]])
    R = t3.f_mean(f)
    expected = np.asarray(
        [0.83, 0.63])
    assert np.allclose(R, expected)


def test_f_range():
    """Tests the f_range fn"""
    f = np.asarray([
        [0.99, 1.0, 0.5],
        [0.69, 0.6, 0.6]])
    R = t3.f_range(f)
    expected = np.asarray(
        [0.5, 0.09])
    assert np.allclose(R, expected)


def test_f_sum():
    """Tests the f_sum fn"""
    f = np.asarray([
        [0.99, 1.0, 0.5],
        [0.69, 0.6, 0.6]])
    R = t3.f_sum(f)
    expected = np.asarray(
        [0.83, 0.63])
    assert np.allclose(R, expected)


def test_f_w_sum():
    """Tests the f_w_sum fn"""
    f = np.asarray([
        [0.99, 1.0, 0.5],
        [0.69, 0.6, 0.6]])
    weights = np.asarray(
        [0.5, 0.25, 0.25])
    R = t3.f_w_sum(f, weights)
    expected = np.asarray(
        [0.87, 0.645])
    assert np.allclose(R, expected)


def test_f_variance():
    """Tests the f_variance fn"""
    f = np.asarray([
        [0.99, 1.0, 0.5],
        [0.69, 0.6, 0.6]])
    R = t3.f_variance(f)
    expected = np.asarray(
        [0.0817, 0.0027])
    assert np.allclose(R, expected)


def test_f_mean_variance():
    """Tests the f_mean_variance fn"""
    f = np.asarray([
        [0.99, 1.0, 0.5],
        [0.69, 0.6, 0.6]])
    R = t3.f_mean_variance(f)
    expected = np.asarray(
        [1.42320289996384, 1.54948632859709])
    assert np.allclose(R, expected)


def test_f_skew():
    """Tests the f_skew fn"""
    f = np.asarray([
        [0.5, 0.99, 1.0],
        [0.6, 0.61, 0.69]])
    R = t3.f_skew(f, reverse=False)
    expected = np.asarray(
        [0.96, -0.777777777777779])
    assert np.allclose(R, expected)
    R = t3.f_skew(f, reverse=True)
    expected = np.asarray(
        [-0.96, 0.777777777777779])
    assert np.allclose(R, expected)


def test_f_kurtosis():
    """Tests the f_skew fn"""
    f = np.asarray([
        [0.5, 0.52, 0.99, 1.0],
        [0.6, 0.61, 0.69, 1.0]])
    R = t3.f_kurtosis(f)
    expected = np.asarray(
        [1.06382978723404, 5.0])
    assert np.allclose(R, expected)



"""RAPID (Robustness Analysis Producing Intelligent Decisions)"""

import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="rapidrobustness",
    version="0.0.2",
    author="Cameron McPhail",
    author_email="cameron.mcphail@adelaide.edu.au",
    description="Calculations and analyses of robustness.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/cameronmcphail/RAPID",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



"""Contains functions that help analyse robustness values.

This includes functions that help analyse the impact of scenarios on
robustness values and robustness rankings, as well as functions that
help analyse the impact of different robustness metrics on the
robustness values and rankings.

This modeule also contains functions to help visualise these impacts.
"""
from .comparisons import (
    scenarios_similarity,
    R_metric_similarity,
    delta_plot,
    tau_plot
)



"""Tests the Robustness metrics"""

import numpy as np
from .. import common_metrics


def test_maximin():
    """Tests the maximin fn"""
    f = np.asarray([
        [0.99, 1.0, 0.5],
        [0.69, 0.6, 0.6]])
    R = common_metrics.maximin(f, maximise=True)
    expected = np.asarray(
        [0.5, 0.6])
    assert np.allclose(R, expected)
    R = common_metrics.maximin(f, maximise=False)
    expected = np.asarray(
        [-1.0, -0.69])
    assert np.allclose(R, expected)


def test_maximax():
    """Tests the maximax fn"""
    f = np.asarray([
        [0.99, 1.0, 0.5],
        [0.69, 0.6, 0.6]])
    R = common_metrics.maximax(f, maximise=True)
    expected = np.asarray(
        [1.0, 0.69])
    assert np.allclose(R, expected)
    R = common_metrics.maximax(f, maximise=False)
    expected = np.asarray(
        [-0.5, -0.6])
    assert np.allclose(R, expected)


def test_hurwicz():
    """Tests the Hurwicz fn"""
    f = np.asarray([
        [0.99, 1.0, 0.5],
        [0.69, 0.6, 0.6]])
    alpha = 0.75
    R = common_metrics.hurwicz(f, maximise=True, alpha=alpha)
    expected = np.asarray(
        [0.625, 0.6225])
    assert np.allclose(R, expected)
    R = common_metrics.hurwicz(f, maximise=False, alpha=alpha)
    expected = np.asarray(
        [-0.875, -0.6675])
    assert np.allclose(R, expected)
    R = common_metrics.hurwicz(f, maximise=False)
    expected = np.asarray(
        [-0.75, -0.645])
    assert np.allclose(R, expected)


def test_laplace():
    """Tests the Laplace fn"""
    f = np.asarray([
        [0.99, 1.0, 0.5],
        [0.69, 0.6, 0.6]])
    R = common_metrics.laplace(f, maximise=True)
    expected = np.asarray(
        [0.83, 0.63])
    assert np.allclose(R, expected)
    R = common_metrics.laplace(f, maximise=False)
    expected = np.asarray(
        [-0.83, -0.63])
    assert np.allclose(R, expected)


def test_minimax_regret():
    """Tests the minimax_regret fn"""
    f = np.asarray([
        [0.99, 1.0, 0.5],
        [0.69, 0.6, 0.6]])
    R = common_metrics.minimax_regret(f, maximise=True)
    expected = np.asarray(
        [-0.1, -0.4])
    assert np.allclose(R, expected)
    R = common_metrics.minimax_regret(f, maximise=False)
    expected = np.asarray(
        [-0.4, -0.1])
    assert np.allclose(R, expected)


def test_percentile_regret():
    """Tests the percentile_regret fn"""
    f = np.asarray([
        [0.99, 1.0, 0.5],
        [0.69, 0.6, 0.6]])
    R = common_metrics.percentile_regret(f, maximise=True, percentile=0.5)
    expected = np.asarray(
        [0.0, -0.3])
    assert np.allclose(R, expected)
    R = common_metrics.percentile_regret(f, maximise=False, percentile=0.5)
    expected = np.asarray(
        [-0.3, 0.0])
    assert np.allclose(R, expected)


def test_mean_variance():
    """Tests the mean_variance fn"""
    f = np.asarray([
        [0.99, 1.0, 0.5],
        [0.69, 0.6, 0.6]])
    R = common_metrics.mean_variance(f, maximise=True)
    expected = np.asarray(
        [1.42320289996384, 1.54948632859709])
    assert np.allclose(R, expected)
    R = common_metrics.mean_variance(f, maximise=False)
    expected = np.asarray(
        [0.132210105461122, 0.351723890540445])
    assert np.allclose(R, expected)


def test_undesirable_deviations():
    """Tests the undesirable_deviations fn"""
    f = np.asarray([
        [0.99, 1.0, 0.5],
        [0.69, 0.6, 0.6]])
    R = common_metrics.undesirable_deviations(f, maximise=True)
    expected = np.asarray(
        [-0.245, 0.0])
    assert np.allclose(R, expected)
    R = common_metrics.undesirable_deviations(f, maximise=False)
    expected = np.asarray(
        [-0.005, -0.045])
    assert np.allclose(R, expected)


def test_percentile_skew():
    """Tests the percentile skew fn"""
    f = np.asarray([
        [0.99, 1.0, 0.5],
        [0.69, 0.6, 0.61]])
    R = common_metrics.percentile_skew(f, maximise=True)
    expected = np.asarray(
        [0.96, -0.777777777777779])
    assert np.allclose(R, expected)
    R = common_metrics.percentile_skew(f, maximise=False)
    expected = np.asarray(
        [-0.96, 0.777777777777779])
    assert np.allclose(R, expected)


def test_percentile_kurtosis():
    """Tests the percentile kurtosis fn"""
    f = np.asarray([
        [0.99, 1.0, 0.5, 0.52],
        [0.69, 0.6, 0.61, 1.0]])
    R = common_metrics.percentile_kurtosis(f, maximise=True)
    expected = np.asarray(
        [1.06382979, 5.0])
    assert np.allclose(R, expected)


def test_starrs_domain():
    """Tests the starrs_domain fn"""
    f = np.asarray([
        [0.5, 0.99, 1.0],
        [0.6, 0.65, 0.69]])
    threshold = 0.65
    R = common_metrics.starrs_domain(
        f,
        maximise=True,
        threshold=threshold,
        accept_equal=True)
    expected = np.asarray(
        [2./3., 2./3.])
    assert np.allclose(R, expected)
    R = common_metrics.starrs_domain(
        f,
        maximise=True,
        threshold=threshold,
        accept_equal=False)
    expected = np.asarray(
        [2./3., 1./3.])
    assert np.allclose(R, expected)
    R = common_metrics.starrs_domain(
        f,
        maximise=False,
        threshold=threshold,
        accept_equal=True)
    expected = np.asarray(
        [1./3., 2./3.])
    assert np.allclose(R, expected)
    R = common_metrics.starrs_domain(
        f,
        maximise=False,
        threshold=threshold,
        accept_equal=False)
    expected = np.asarray(
        [1./3., 1./3.])
    assert np.allclose(R, expected)
